@@ -51,7 +51,18 @@ start_component (void)
     engines = ibus_component_get_engines (component);
     for (p = engines; p != NULL; p = p->next) {
         IBusEngineDesc *engine = (IBusEngineDesc *)p->data;
-        ibus_factory_add_engine (factory, engine->name, IBUS_TYPE_M17N_ENGINE);
+#if IBUS_CHECK_VERSION(1,3,99)
+        const gchar *engine_name = ibus_engine_desc_get_name (engine);
+#else
+        const gchar *engine_name = engine->name;
+#endif  /* !IBUS_CHECK_VERSION(1,3,99) */
+        GType type = ibus_m17n_engine_get_type_for_name (engine_name);
+
+        if (type == G_TYPE_INVALID) {
+            g_debug ("Can not create engine type for %s", engine_name);
+            continue;
+        }
+        ibus_factory_add_engine (factory, engine_name, type);
     }
 
     if (ibus) {
@@ -74,8 +85,7 @@ print_engines_xml (void)
 
     ibus_init ();
 
-    bus = ibus_bus_new ();
-    ibus_m17n_init (bus);
+    ibus_m17n_init_common ();
 
     component = ibus_m17n_get_component ();
     output = g_string_new ("");
