@@ -83,6 +83,7 @@ parse_value (MPlist *plist, gchar *text)
     if (mplist_key (plist) == Minteger) {
         long val;
 
+        errno = 0;
         val = strtol (text, NULL, 10);
         if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN))
             || (errno != 0 && val == 0))
@@ -209,10 +210,8 @@ color_to_gdk (guint color, GdkColor *color_gdk)
 static void
 set_color (ConfigContext *context, const gchar *name, GdkColor *color)
 {
-    GValue value = { 0 };
     gchar buf[8];
 
-    g_value_init (&value, G_TYPE_STRING);
     if (color)
         sprintf (buf, "#%02X%02X%02X",
                  (color->red & 0xFF00) >> 8,
@@ -220,8 +219,7 @@ set_color (ConfigContext *context, const gchar *name, GdkColor *color)
                  (color->blue & 0xFF00) >> 8);
     else
         strcpy (buf, "none");
-    g_value_set_string (&value, buf);
-    ibus_config_set_value (config, context->section, name, &value);
+    ibus_m17n_config_set_string (config, context->section, name, buf);
 }
 
 static void
@@ -253,17 +251,16 @@ on_underline_changed (GtkComboBox *combo,
     ConfigContext *context = user_data;
     GtkTreeModel *model;
     GtkTreeIter iter;
-    GValue value = { 0 };
     gint active;
 
     model = gtk_combo_box_get_model (combo);
     gtk_combo_box_get_active_iter (combo, &iter);
     gtk_tree_model_get (model, &iter, COLUMN_VALUE, &active, -1);
 
-    g_value_init (&value, G_TYPE_INT);
-    g_value_set_int (&value, active);
-    ibus_config_set_value (config, context->section, "preedit_underline",
-                           &value);
+    ibus_m17n_config_set_int (config,
+                              context->section,
+                              "preedit_underline",
+                              active);
 }
 
 static void
@@ -273,17 +270,16 @@ on_orientation_changed (GtkComboBox *combo,
     ConfigContext *context = user_data;
     GtkTreeModel *model;
     GtkTreeIter iter;
-    GValue value = { 0 };
     gint active;
 
     model = gtk_combo_box_get_model (combo);
     gtk_combo_box_get_active_iter (combo, &iter);
     gtk_tree_model_get (model, &iter, COLUMN_VALUE, &active, -1);
 
-    g_value_init (&value, G_TYPE_INT);
-    g_value_set_int (&value, active);
-    ibus_config_set_value (config, context->section, "lookup_table_orientation",
-                           &value);
+    ibus_m17n_config_set_int (config,
+                              context->section,
+                              "lookup_table_orientation",
+                              active);
 }
 
 static void
@@ -366,7 +362,7 @@ start (const gchar *engine_name)
     GError *error = NULL;
     GtkCellRenderer *renderer;
     ConfigContext context;
-    GValue value = { 0 };
+    gchar *color;
     gboolean is_foreground_set, is_background_set;
     GdkColor foreground, background;
     gint underline;
@@ -416,15 +412,14 @@ start (const gchar *engine_name)
     /* foreground color of pre-edit buffer */
     is_foreground_set = FALSE;
     color_to_gdk (PREEDIT_FOREGROUND, &foreground);
-    if (ibus_config_get_value (config, context.section, "preedit_foreground",
-                               &value)) {
-        const gchar *color;
-
-        color = g_value_get_string (&value);
+    if (ibus_m17n_config_get_string (config,
+                                     context.section,
+                                     "preedit_foreground",
+                                     &color)) {
         if (g_strcmp0 (color, "none") != 0 &&
             gdk_color_parse (color, &foreground))
             is_foreground_set = TRUE;
-        g_value_unset (&value);
+        g_free (color);
     }
 
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(checkbutton_foreground),
@@ -444,16 +439,15 @@ start (const gchar *engine_name)
     /* background color of pre-edit buffer */
     is_background_set = FALSE;
     color_to_gdk (PREEDIT_BACKGROUND, &background);
-    if (ibus_config_get_value (config, context.section, "preedit_background",
-                               &value)) {
-        const gchar *color;
-
-        color = g_value_get_string (&value);
+    if (ibus_m17n_config_get_string (config,
+                                     context.section,
+                                     "preedit_background",
+                                     &color)) {
         if (g_strcmp0 (color, "none") != 0 &&
             gdk_color_parse (color, &background))
             is_background_set = TRUE;
         g_debug ("preedit_background %d", is_background_set);
-        g_value_unset (&value);
+        g_free (color);
     }
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(checkbutton_background),
                                   is_background_set);
@@ -474,12 +468,11 @@ start (const gchar *engine_name)
                                 renderer, TRUE);
     gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT(combobox_underline),
                                     renderer, "text", 0, NULL);
-    underline = IBUS_ATTR_UNDERLINE_NONE;
-    if (ibus_config_get_value (config, context.section, "preedit_underline",
-                               &value)) {
-        underline = g_value_get_int (&value);
-        g_value_unset (&value);
-    }
+    if (!ibus_m17n_config_get_int (config,
+                                   context.section,
+                                   "preedit_underline",
+                                   &underline))
+        underline = IBUS_ATTR_UNDERLINE_NONE;
 
     index = get_combo_box_index_by_value (GTK_COMBO_BOX(combobox_underline),
                                               underline);
@@ -493,13 +486,11 @@ start (const gchar *engine_name)
                                 renderer, TRUE);
     gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT(combobox_orientation),
                                     renderer, "text", 0, NULL);
-    orientation = IBUS_ORIENTATION_SYSTEM;
-    if (ibus_config_get_value (config, context.section,
-                               "lookup_table_orientation",
-                               &value)) {
-        orientation = g_value_get_int (&value);
-        g_value_unset (&value);
-    }
+    if (!ibus_m17n_config_get_int (config,
+                                   context.section,
+                                   "lookup_table_orientation",
+                                   &orientation))
+        orientation = IBUS_ORIENTATION_SYSTEM;
 
     index = get_combo_box_index_by_value (GTK_COMBO_BOX(combobox_orientation),
                                           orientation);
